from typing import Optional, List

from zpy.api.http.status_codes import HttpStatus


class ZError(Exception):
    """Base Error

    Args:
        Exception ([type]): [description]
    """

    def __init__(
            self, message: str, reason: str, details: Optional[List[str]], meta: Optional[dict] = None,
            parent_ex: Exception = None, *args: object
    ) -> None:
        super().__init__(message, *args)
        self.reason = reason
        self.message = message
        self.details = details
        self.metadata = meta
        self.internal_exception = parent_ex

    def add_detail(self, message: str) -> None:
        if self.details is None:
            self.details = []
        self.details.append(message)

    def set_metadata(self, value: dict):
        self.metadata = value

    def __str__(self):
        return f'[ZCE] - {self.message}\n\t  - {self.reason}'


class ZHttpError(ZError):
    """Http Base Error

    Args:
        ZError ([type]): [description]
    """

    def __init__(
            self,
            message: str = None,
            reason: str = None,
            details: Optional[List[str]] = None,
            status: HttpStatus = HttpStatus.INTERNAL_SERVER_ERROR,
            meta: Optional[dict] = None,
            parent_ex: Exception = None,
            *args: object
    ) -> None:
        super().__init__(message, reason, details, meta, parent_ex, *args)
        self.status = status


class BadRequest(ZHttpError):
    """BadRequest

    Args:
        ZError ([type]): [description]
    """

    def __init__(
            self,
            message: str = None,
            reason: str = None,
            details: Optional[List[str]] = None,
            meta: Optional[dict] = None,
            parent_ex: Exception = None,
            *args: object
    ) -> None:
        super().__init__(message, reason, details, HttpStatus.BAD_REQUEST, meta, parent_ex, *args)


class Unauthorized(ZHttpError):
    """Unauthorized

    Args:
        ZError ([type]): [description]
    """

    def __init__(
            self,
            message: str = None,
            reason: str = None,
            details: Optional[List[str]] = None,
            meta: Optional[dict] = None,
            parent_ex: Exception = None,
            *args: object
    ) -> None:
        super().__init__(message, reason, details, HttpStatus.UNAUTHORIZED, meta, parent_ex, *args)


class NotFound(ZHttpError):
    """BadRequest

    Args:
        ZError ([type]): [description]
    """

    def __init__(
            self,
            message: str = None,
            reason: str = None,
            details: Optional[List[str]] = None,
            meta: Optional[dict] = None,
            parent_ex: Exception = None,
            *args: object
    ) -> None:
        super().__init__(message, reason, details, HttpStatus.NOT_FOUND, meta, parent_ex, *args)


class UnprocessableEntity(ZHttpError):
    """UnprocessableEntity

    Args:
        ZError ([type]): [description]
    """

    def __init__(
            self,
            message: str = None,
            reason: str = None,
            details: Optional[List[str]] = None,
            meta: Optional[dict] = None,
            parent_ex: Exception = None,
            *args: object
    ) -> None:
        super().__init__(message, reason, details, HttpStatus.UNPROCESSABLE, meta, parent_ex, *args)



from typing import Any, Callable, List, Optional
from functools import reduce
import logging


def safely_exec(callable_fn: Callable, args=None) -> Optional[Any]:
    """
    Execute provided function in try:except block
    @param callable_fn:
    @param args:
    @return: value returned of execution or none
    """
    if args is None:
        args = []
    try:
        return callable_fn(*args)
    except Exception as e:
        logging.exception(e)
    return None


def exec_if_nt_null(callable_fn: Callable, args: Optional[List[Any]] = None) -> object:
    """
    Execute function if args not null
    """
    if args is None:
        args = []
    for arg in args:
        if arg is None:
            return False
    return callable_fn(*args)


def safely_exec_with(callable_fn: Callable, default_value: Any = None, args=None) -> Optional[Any]:
    """
    Execute provided function in try:except block
    @param default_value:
    @param callable_fn:
    @param args:
    @return: value returned of execution or none
    """
    if args is None:
        args = []
    try:
        return callable_fn(*args)
    except Exception as e:
        logging.exception(e)
    return default_value


def fn_composite(*func):
    """
    Function composition
    @param func: functions
    @return: composition
    """

    def compose(f, g):
        return lambda x: f(g(x))

    return reduce(compose, func, lambda x: x)



from abc import ABC, abstractmethod

from flask import Flask


class ZHook(ABC):
    @abstractmethod
    def execute(self, app: Flask, **kwargs):
        pass


class NotFoundHook(ZHook):
    def execute(self, app: Flask, **kwargs):
        @app.errorhandler(404)
        def not_found(e):
            return "Resource requested not found", 404

        return super().execute(app)



# Created by Noé Cruz | Zurckz 22 at 05/03/2022
# See https://www.linkedin.com/in/zurckz
import logging
from typing import Optional
from zpy.utils import get_env

from zpy.logger import zL, ZLFormat, ZLogger


class ZContext(object):

    def __init__(self, logger: Optional[ZLogger] = None):
        self.logger = logger
        if not logger:
            self.logger = zL(get_env("APP_NAME", "API"), logging.INFO, ZLFormat.LM)

    def release_logger(self, name: str = "API"):
        self.logger.release(name)


context: Optional[ZContext] = None


def setup_context(ctx: Optional[ZContext] = None) -> ZContext:
    global context
    if ctx:
        context = ctx
    if not ctx:
        context = ZContext()
    return context


def zapp_context() -> ZContext:
    global context
    if not context:
        return setup_context()
    return context



basic = {
    "src": {
        "type": "dir",
        "name": "src",
        "child": {
            "api": {
                "type": "dir",
                "name": "api",
                "child": {
                    "routes.py": {
                        "type": "file",
                        "name": "routes.py",
                        "content": """# 🛸 Generated by zPy
from di import @usecase_var
from flask import Flask
from zpy.api.http.response import response_builder
from zpy.api.flask import create_app


app: Flask = create_app()


@app.route("/api/awesome", methods=["GET"])
@response_builder()
def awesome_api():
    return @usecase_var.execute(None)"""
                    }
                }
            },
            "contexts": {
                "type": "dir",
                "name": "contexts",
                "child": {
                    "@context_name": {
                        "type": "dir",
                        "name": "@context_name",
                        "child": {
                            "application": {
                                "type": "dir",
                                "name": "application",
                                "child": {
                                    "@usecase_file.py": {
                                        "type": "file",
                                        "name": "@usecase_file.py",
                                        "content": """# 🛸 Generated by zPy
from typing import Any

from zpy.app.usecase import UseCase


from ..domain.repositories import @repository_name


class @usecase_class(UseCase[Any, Any]):
    \"\"\"
        Use Case description.
    \"\"\"

    def __init__(self, repository: @repository_name) -> None:
        self.repository = repository

    def execute(self, data: Any, *args, **kwargs) -> None:
        # TODO Do magic with business rules 😁
        return self.repository.@repository_action(data)"""
                                    }
                                }

                            },
                            "domain": {
                                "type": "dir",
                                "name": "domain",
                                "child": {
                                    "catalogs.py": {
                                        "type": "file",
                                        "name": "catalogs.py",
                                        "content": """# 🛸 Generated by zPy
from enum import Enum

class SomeCatalog(Enum):
    \"\"\"
        Catalog example
    \"\"\"
    SOME_CONSTANT = 9
    ANOTHER_CONSTANT = 10"""
                                    },
                                    "repositories.py": {
                                        "type": "file",
                                        "name": "repositories.py",
                                        "content": """# 🛸 Generated by zPy
from abc import abstractmethod
from abc import ABC
from typing import Any


class @repository_name(ABC):
    @abstractmethod
    def @repository_action(self, data: Any) -> Any:
        \"\"\"
        Description

        :param data: input data
        :return: result
        \"\"\"
        ..."""
                                    }
                                }

                            },
                            "infraestructure": {
                                "type": "dir",
                                "name": "infraestructure",
                                "child": {
                                    "@repository_impl_file.py": {
                                        "type": "file",
                                        "name": "@repository_file.py",
                                        "content": """# 🛸 Generated by zPy
from ..domain.repositories import @repository_name
from typing import Any

class @repository_impl(@repository_name):

    def __init__(self):
        ...

    def @repository_action(self, data: Any) -> Any:
        try:
            # TODO Do magic with data 😁
            return "😁"
        finally:
            ..."""
                                    }
                                }

                            },
                            "__init__.py": {
                                "type": "file",
                                "name": "__init__.py",
                                "content": """# 🛸 Generated by zPy
# 🛰 Application Layer

from .application.@usecase_file import @usecase_class

# 📦  Domain Layer

from .domain.repositories import @repository_name


# Infrastructure

from .infraestructure.@repository_file import @repository_impl"""
                            }
                        }
                    }
                }
            },
            "di.py": {
                "type": "file",
                "name": "di.py",
                "content": """# 🛸 Generated by zPy
from zpy.utils import get_env_or_throw as var
from zpy.app.usecase import UseCase
from typing import Any

# * Setup Dependencies 📃
from contexts.@context_name.domain.repositories import @repository_name
from contexts.@context_name.infraestructure.@repository_file import @repository_impl
from contexts.@context_name.application.@usecase_file import @usecase_class

repository: @repository_name = @repository_impl()


# Setup UseCases

@usecase_var: UseCase[Any, None] = @usecase_class(repository)
print("🚀 Dependencies loaded successfully...")"""
            },
            "handler.py": {
                "type": "file",
                "name": "handler.py",
                "content": """# 🛸 Generated by zPy
from zpy.api.flask.cloud_handlers import aws_lambda

from api.routes import app
import aws_lambda_wsgi as awsgi


@aws_lambda()
def handle(event: dict, context: dict) -> any:
    return awsgi.response(app, event, context)"""
            },
            "local_deploy.py": {
                "type": "file",
                "name": "local_deploy.py",
                "content": """# 🛸 Generated by zPy
from dotenv import load_dotenv
load_dotenv(
    dotenv_path="@base_path\\\\.env"
)

from api.routes import app

if __name__ == "__main__":
    app.run(host="127.0.0.1", port="5050", debug=True)"""
            }
        }
    },
    "tests": {
        "type": "dir",
        "name": "tests",
        "child": {
            "@usecase_file_test.py": {
                "type": "file",
                "name": "@usecase_file_test.py",
                "content": """# 🛸 Generated by zPy
from dotenv import load_dotenv
load_dotenv(dotenv_path="..\\.env")

from src.handler import handle

if __name__ == "__main__":
    handle({}, {})"""
            }
        }
    },
    ".env": {
        "type": "file",
        "name": ".env",
        "content": "ENVIRONMENT=local\nORACLE_CLIENT_PATH=C:\oracle\win32\instantclient_19_9"
    },
    ".gitignore": {
        "type": "raw-file",
        "name": ".gitignore",
        "path": "ignore.py"
    },
    "package.json": {
        "type": "file",
        "content": """{
    "name": "@project_name",
    "version": "1.0.0"
}""",
        "name": "package.json"
    },
    "CHANGELOG.md": {
        "type": "file",
        "name": "CHANGELOG.md",
        "content": """# Change Log
All notable changes to this project will be documented in this file.

The format is based on [Keep a Changelog](http://keepachangelog.com/)
and this project adheres to [Semantic Versioning](http://semver.org/).

## [1.0.0] - @date Project scaffolding generated by @project_name

> Author: @developer | @dev_url

### Added

- Project scaffolding configured
- Default Use Case implemented
- Default Repository Implementation"""
    },
    "README.md": {
        "type": "file",
        "name": "README.md",
        "content": """# @project_name

@project_description

## Getting Started

Clone this repository for local deploy.

### Prerequisites

- python3.9
- pip
- virtualenv

### Dependencies

Some dependencies are optional according of the requirements.
- zpy
- zdb
- boto3

For more information see requirements.txt file

### Installing

- Create virtual environment, activate it and download all dependencies.

```
  virtualenv venv
```

- Clone repository

```

git clone http-url-repository

```

## Running the tests

Load env vars

```bash
ENVIRONMENT=
```

Test use cases

```python

from src.di import @usecase_class

if __name__ == "__main__":
    @usecase_var.execute(@usecase_repository_dependency)


```

Or test from handler

`````python
from dotenv import load_dotenv

load_dotenv()

from src.handler import handle

if __name__ == "__main__":
    handle({}, {})

`````

Use selenium or cucumber

## Built With

- Python 3.9
- zpy cli

## Versioning

We use [SemVer](http://semver.org/) for versioning. For the versions available, see
the [tags on this repository](./CHANGELOG.md)
.

## Authors

- **@developer** - _[Backend Developer](@dev_url)_
"""
    },
    "requirements.txt": {
        "type": "file",
        "name": "requirements.txt",
        "content":
            """zpy-api-core
python-dotenv
awsgi"""
    }
}



from dataclasses import is_dataclass, asdict
from typing import Any, Union
import json
from zpy.utils.objects import ZObjectModel
from zpy.utils.values import is_built_type


def serialize_object_value(value: Any) -> Union[Union[dict, str], Any]:
    if issubclass(value.__class__, ZObjectModel):
        return value.sdump()

    if is_dataclass(value):
        return asdict(value)

    if is_built_type(value) is False:
        try:
            return value.__dict__
        except:
            return value
    else:
        try:
            if isinstance(value, dict):
                return value
            if isinstance(value, tuple):
                return value
            if isinstance(value, list):
                f = []
                for i in value:
                    s = serialize_object_value(i)
                    f.append(s)
                #a = json.dumps(f)
                return f
            return json.dumps(value)
        except Exception as e:
            return value



# Created by Noé Cruz | Zurckz 22 at 20/03/2022
# See https://www.linkedin.com/in/zurckz
from typing import Callable, Optional, Any, Union
from zpy.app import zapp_context as ctx
import json
import copy


class SQSHandler(object):

    def __init__(self, processor: Optional[Callable[[dict, dict], Any]], single: Optional[bool] = False,
                 strict=False, verbose: bool = False, jsonfy=True, re_enqueue: bool = False):
        self.processor = processor
        self.single_item = single
        self.strict = strict
        self.verbose = verbose
        self.response_builder = None
        self.request_parser = None
        self.error_notifier = None
        self.starting_precess_logger = None
        self.jsonfy = jsonfy
        self.re_enqueue = re_enqueue

    @classmethod
    def of(cls, processor: Optional[Callable[[dict, dict], Any]], single: Optional[bool] = False, strict=False,
           verbose=False, jsonfy=True):
        return cls(processor, single, strict, verbose, jsonfy)

    @staticmethod
    def proxy(event, handler):
        return handler.run(event)

    def __notify_error(self, record, msg):
        if self.error_notifier:
            self.error_notifier(record, msg)

    def run(self, event):
        logger = ctx().logger
        failed = False

        if self.request_parser:
            event = self.request_parser(event)

        if not self.processor:
            msg = "The sqs record processor cant be null."
            self.__notify_error(None, msg)
            return msg

        if 'Records' not in event:
            if self.strict:
                msg = "Fatal :: Could not continue processing, invalid event!"
                self.__notify_error(event, msg)
                logger.err(msg)
                return msg
            event = {'Records': [event]}
        batch_results = []
        for index, record in enumerate(event["Records"]):
            body: Optional[Union[dict, str]] = None
            x_record = copy.copy(record)
            record_id = index if 'messageId' not in record else record['messageId']
            try:
                if 'body' not in record:
                    msg = "Fatal :: Invalid item, could not continue processing..."
                    self.__notify_error(x_record, msg)
                    if self.strict:
                        if self.single_item:
                            return msg
                        continue
                    body = record
                else:
                    body = record.pop('body')
                if type(body) is not dict:
                    if self.jsonfy:
                        body = json.loads(body)
                if self.verbose:
                    logger.info(f"Starting record processing: {record_id} with data: {body}")
                if self.starting_precess_logger:
                    self.starting_precess_logger(body, record)
                result = self.processor(body, record)
                if self.verbose:
                    logger.info(result)
                batch_results.append({
                    "result": result,
                    "message_id": record_id
                })
                if self.single_item:
                    break
            except Exception as e:
                failed = True
                self.__notify_error(x_record, "Fatal :: An error occurred while processing notification data")
                logger.exception(f"An error occurred while processing notification data: {body}\n", exc_info=e)
                if self.re_enqueue:
                    raise Exception("Exception autogenerated for re enqueue record...")
        response = {
            'message': "Finished Process." if not failed else "Finished process with errors.",
            "results": batch_results
        }
        if self.response_builder:
            response = self.response_builder(response)
        return response

    def single(self, body, context):
        ...



from typing import Any, List, Optional, Dict
from flask_cors import CORS

from flask import Flask

from zpy.api.flask.zhooks import ZHook
from zpy.api.flask.zmiddlewares import ZMiddleware


def create_flask_app(
        config: dict,
        path_cors_allow: str = None,
        origins: List[str] = None
) -> Flask:
    """
    Flask app builder
    """
    app = Flask(__name__, instance_relative_config=True)
    app.config.from_object(config)
    path_allow = "/*" if path_cors_allow is None or path_cors_allow == '' else path_cors_allow
    _origins = '*' if origins is None or len(origins) <= 0 else origins
    CORS(app, resources={path_allow: {"origins": _origins}})
    return app


def create_app(app: Optional[Flask] = None,
               mw: Optional[List[ZMiddleware]] = None,
               mw_args: Optional[List[Any]] = None,
               hk: Optional[List[ZHook]] = None,
               hk_args: Optional[List[Any]] = None,
               shared_data: Dict[Any, Any] = None,
               path_cors_allow: str = None,
               origins_cors: List[str] = None
               ) -> Flask:
    """
    API App Builder
    @param app: Flask application instance.
    @param mw:  Middlewares
    @param mw_args:  Middleware arguments
    @param hk: Hooks
    @param hk_args: hooks arguments
    @param shared_data: data for setup in flask context
    @param path_cors_allow: Path for configure cors origins. Default: '/*'
    @param origins_cors: Origins for configure cors. Default '*'
    @return:  flask instance
    """
    if hk_args is None:
        hk_args = []
    if mw_args is None:
        mw_args = []
    if shared_data is None:
        shared_data = {}
    if hk is None:
        hk = []
    if mw is None:
        mw = []
    if app is None:
        app = create_flask_app(shared_data, path_cors_allow, origins_cors)

    for i, m in enumerate(mw):
        args = mw_args[i] if i < len(mw_args) else {}
        args.update(shared_data)
        app.wsgi_app = m(app.wsgi_app, **args)
    for i, h in enumerate(hk):
        args = hk_args[i] if i < len(hk_args) else {}
        args.update(shared_data)
        h().execute(app, **args)

    return app



from typing import Optional, List, Callable, Union
from zpy.api.http.errors import BadRequest
from zpy.utils.objects import ZObjectModel

__author__ = "Noé Cruz | Zurck'z 2021"
__version__ = "1.0.0"


class ZHeader:
    """
    Zurck'z Header Wrapper
    """

    def __init__(self, raw_headers: dict):
        self.headers = raw_headers

    def to_model(self, model: Union[Callable, ZObjectModel]) -> ZObjectModel:
        return model(**self.headers)

    def to_dict(self) -> dict:
        final_headers = self.headers
        return final_headers

    @classmethod
    def from_dict(cls, raw: dict):
        return ZHeader(raw)


def get_headers_or_raise(
    request: dict,
    headers: List[str],
    throw: Optional[bool] = True,
    check_value: Optional[bool] = True,
    include_all: Optional[bool] = False,
) -> ZHeader:
    """
    Check if keys provided as headers are in the headers dict request
    :param request: headers
    :param throw : Raise exception if some header not found
    :param check_value: Verify header are in request and value is different of null or empty
    :param include_all: return all headers
    :param headers: headers key to validate
    :return: dict with headers
    """
    validated_headers: dict = {}
    for header in headers:
        header_value: Optional[str] = None
        if header in request:
            header_value = request[header]
            if check_value:
                if header_value is None or header_value.strip() == "" and throw:
                    raise BadRequest(
                        message=f"The value of header: '{header}' cannot be null or empty"
                    )
        elif throw:
            raise BadRequest(message=f"The header: '{header}' is missing")
        validated_headers[header] = header_value

    if include_all:
        request.update(validated_headers)
        return ZHeader.from_dict(request)

    return ZHeader.from_dict(validated_headers)



from typing import List, Dict

from zpy.cli.agregations import add_config, add_context, add_use_case
from zpy.cli.scaffolding import scaffolding_dispatcher

import click


@click.group()
def cli():
    """This is an example script to learn Click."""
    ...


@cli.command()
@click.option('--project', '-p', required=True, type=str, help='Project name')
@click.option('--verbose', '-v', required=False, is_flag=True, default=False, help='Show logs')
@click.option('--description', '-d', required=False, type=str, help='Project description',
              default="This is very boring haha xd...")
@click.option('--developer', '-dev', required=False, type=str, help='Author', default="Zurck'z 2022")
@click.option('--context', '-c', required=False, type=str, help='Context name', default="Users")
@click.option('--use_case', '-uc', required=False, type=str, help='Use Case Name', default="GetUser")
@click.option('--developer_url', '-dev-url', required=False, type=str, help='Author url',
              default="https://www.linkedin.com/in/zurckz/")
@click.option('--open_project', '-op', required=False, is_flag=True, default=False, help='Open project')
@click.option('--only_content', '-oc', required=False, is_flag=True, default=False,
              help='Create only structure project')
def make(project, verbose, description, developer, context, use_case, developer_url, open_project,
         only_content):
    in_data: List[Dict[str, str]] = [{"value": project, "title": "Project name:"},
                                     {"value": description, "title": "Project description:"},
                                     {"value": developer, "title": "Author:"},
                                     {"value": developer_url, "title": "Url Author:"},
                                     {"value": context.lower(), "title": "Context:"},
                                     {"value": use_case, "title": "Use Case:"},
                                     {"value": context.title(), "title": "Repository"},
                                     {"value": open_project, "title": "Open Project?:"},
                                     {"value": only_content, "title": "Only Content?:"},
                                     None
                                     ]
    if verbose is True:
        click.echo(f'\N{grinning face} - Initializing for create new project...')
        for data in in_data:
            if data is not None:
                click.echo(f'\N{grinning face} - With {data["title"]} {data["value"]}')

    scaffolding_dispatcher(in_data, verbose, click)


@cli.group()
def add():
    ...


@add.command()
@click.option('--name', '-n', required=True, type=str, is_flag=False, help='Bounded Context name')
@click.option('--use_case', '-uc', required=False, type=str, help='Use Case Name', default="GetUser")
@click.option('--force', '-f', required=False, type=bool, is_flag=True, help='Force creation, skipping project',
              default=False)
def context(name: str, use_case: str, force: bool):
    add_context(name, use_case, force)


@add.command()
@click.option('--name', '-n', required=True, type=str, is_flag=False, help='Use case name')
@click.option('--context', '-c', required=True, type=str, is_flag=False,
              help='Context name where use case will be created')
# @click.option('--repository', '-r', required=False, type=str, help='Repository dependency', default="AwesomeRepository")
def case(name: str, context: str):
    add_use_case(name, context)


@add.command()
@click.option('--database', '-db', required=True, type=bool, is_flag=True, help='Database configuration')
def config(database: bool):
    add_config(database)


@cli.command()
def drop():
    click.echo('Dropping...')



ignore: str = """# OS file control
**desktop.ini
**.DS_Store
.vscode
install_zpy.bat
run.bat
run.py
tests.py
resources
src/local_deploy.py
# AWS
/.aws-sam/

# Compiled class file
*.class

# Log file
*.log

# BlueJ files
*.ctxt

# Mobile Tools for Java (J2ME)
.mtj.tmp/

# Package Files #
*.jar
*.war
*.nar
*.ear
*.zip
*.tar.gz
*.rar

# virtual machine crash logs, see http://www.java.com/en/download/help/error_hotspot.xml
hs_err_pid*

# Xcode
#
# gitignore contributors: remember to update Global/Xcode.gitignore, Objective-C.gitignore & Swift.gitignore

## User settings
xcuserdata/

## compatibility with Xcode 8 and earlier (ignoring not required starting Xcode 9)
*.xcscmblueprint
*.xccheckout

## compatibility with Xcode 3 and earlier (ignoring not required starting Xcode 4)
build/
DerivedData/
*.moved-aside
*.pbxuser
!default.pbxuser
*.mode1v3
!default.mode1v3
*.mode2v3
!default.mode2v3
*.perspectivev3
!default.perspectivev3

## Obj-C/Swift specific
*.hmap

## App packaging
*.ipa
*.dSYM.zip
*.dSYM

## Playgrounds
timeline.xctimeline
playground.xcworkspace

# Swift Package Manager
#
# Add this line if you want to avoid checking in source code from Swift Package Manager dependencies.
# Packages/
# Package.pins
# Package.resolved
# *.xcodeproj
#
# Xcode automatically generates this directory with a .xcworkspacedata file and xcuserdata
# hence it is not needed unless you have added a package configuration file to your project
# .swiftpm

.build/

# CocoaPods
#
# We recommend against adding the Pods directory to your .gitignore. However
# you should judge for yourself, the pros and cons are mentioned at:
# https://guides.cocoapods.org/using/using-cocoapods.html#should-i-check-the-pods-directory-into-source-control
#
# Pods/
#
# Add this line if you want to avoid checking in source code from the Xcode workspace
# *.xcworkspace

# Carthage
#
# Add this line if you want to avoid checking in source code from Carthage dependencies.
# Carthage/Checkouts

Carthage/Build/

# Accio dependency management
Dependencies/
.accio/

# fastlane
#
# It is recommended to not store the screenshots in the git repo.
# Instead, use fastlane to re-generate the screenshots whenever they are needed.
# For more information about the recommended setup visit:
# https://docs.fastlane.tools/best-practices/source-control/#source-control

fastlane/report.xml
fastlane/Preview.html
fastlane/screenshots/**/*.png
fastlane/test_output

# Code Injection
#
# After new code Injection tools there's a generated folder /iOSInjectionProject
# https://github.com/johnno1962/injectionforxcode

iOSInjectionProject/

# Byte-compiled / optimized / DLL files
__pycache__/
*.py[cod]
*$py.class

# C extensions
*.so

# Distribution / packaging
.Python
build/
develop-eggs/
dist/
downloads/
eggs/
.eggs/
lib/
lib64/
parts/
sdist/
var/
wheels/
share/python-wheels/
*.egg-info/
.installed.cfg
*.egg
MANIFEST

# PyInstaller
#  Usually these files are written by a python script from a template
#  before PyInstaller builds the exe, so as to inject date/other infos into it.
*.manifest
*.spec

# Installer logs
pip-log.txt
pip-delete-this-directory.txt

# Unit tests / coverage reports
htmlcov/
.tox/
.nox/
.coverage
.coverage.*
.cache
nosetests.xml
coverage.xml
*.cover
*.py,cover
.hypothesis/
.pytest_cache/
cover/

# Translations
*.mo
*.pot

# Django stuff:
*.log
local_settings.py
db.sqlite3
db.sqlite3-journal

# Flask stuff:
instance/
.webassets-cache

# Scrapy stuff:
.scrapy

# Sphinx documentation
docs/_build/

# PyBuilder
.pybuilder/
target/

# Jupyter Notebook
.ipynb_checkpoints

# IPython
profile_default/
ipython_config.py

# pyenv
#   For a library or package, you might want to ignore these files since the code is
#   intended to run in multiple environments; otherwise, check them in:
# .python-version

# pipenv
#   According to pypa/pipenv#598, it is recommended to include Pipfile.lock in version control.
#   However, in case of collaboration, if having platform-specific dependencies or dependencies
#   having no cross-platform support, pipenv may install dependencies that don't work, or not
#   install all needed dependencies.
#Pipfile.lock

# PEP 582; used by e.g. github.com/David-OConnor/pyflow
__pypackages__/

# Celery stuff
celerybeat-schedule
celerybeat.pid

# SageMath parsed files
*.sage.py

# Environments
.env
.venv
env/
venv/
ENV/
env.bak/
venv.bak/

# Spyder project settings
.spyderproject
.spyproject

# Rope project settings
.ropeproject

# mkdocs documentation
/site

# mypy
.mypy_cache/
.dmypy.json
dmypy.json

# Pyre type checker
.pyre/

# pytype static type analyzer
.pytype/

# Cython debug symbols
cython_debug/

# Logs
logs
*.log
npm-debug.log*
yarn-debug.log*
yarn-error.log*
lerna-debug.log*

# Diagnostic reports (https://nodejs.org/api/report.html)
report.[0-9]*.[0-9]*.[0-9]*.[0-9]*.json

# Runtime data
pids
*.pid
*.seed
*.pid.lock

# Directory for instrumented libs generated by jscoverage/JSCover
lib-cov

# Coverage directory used by tools like istanbul
coverage
*.lcov

# nyc tests coverage
.nyc_output

# Grunt intermediate storage (https://gruntjs.com/creating-plugins#storing-task-files)
.grunt

# Bower dependency directory (https://bower.io/)
bower_components

# node-waf configuration
.lock-wscript

# Compiled binary addons (https://nodejs.org/api/addons.html)
build/Release

# Dependency directories
node_modules/
jspm_packages/

# Snowpack dependency directory (https://snowpack.dev/)
web_modules/

# TypeScript cache
*.tsbuildinfo

# Optional npm cache directory
.npm

# Optional eslint cache
.eslintcache

# Microbundle cache
.rpt2_cache/
.rts2_cache_cjs/
.rts2_cache_es/
.rts2_cache_umd/

# Optional REPL history
.node_repl_history

# Output of 'npm pack'
*.tgz

# Yarn Integrity file
.yarn-integrity

# dotenv environment variables file
.env
.env.test

# parcel-bundler cache (https://parceljs.org/)
.cache
.parcel-cache

# Next.js build output
.next
out

# Nuxt.js build / generate output
.nuxt
dist

# Gatsby files
.cache/
# Comment in the public line in if your project uses Gatsby and not Next.js
# https://nextjs.org/blog/next-9-1#public-directory-support
# public

# vuepress build output
.vuepress/dist

# Serverless directories
.serverless/

# FuseBox cache
.fusebox/

# DynamoDBService Local files
.dynamodb/

# TernJS port file
.tern-port

# Stores VSCode versions used for testing VSCode extensions
.vscode-test

# yarn v2
.yarn/cache
.yarn/unplugged
.yarn/build-state.yml
.yarn/install-state.gz
.pnp.*

target/
pom.xml.tag
pom.xml.releaseBackup
pom.xml.versionsBackup
pom.xml.next
release.properties
dependency-reduced-pom.xml
buildNumber.properties
.mvn/timing.properties
# https://github.com/takari/maven-wrapper#usage-without-binary-jar
.mvn/wrapper/maven-wrapper.jar

.gradle
**/build/
!src/**/build/

# Ignore Gradle GUI config
gradle-app.setting

# Avoid ignoring Gradle wrapper jar file (.jar files are usually ignored)
!gradle-wrapper.jar

# Cache of project
.gradletasknamecache

# # Work around https://youtrack.jetbrains.com/issue/IDEA-116898
# gradle/wrapper/gradle-wrapper.properties

# Built application files
*.apk
*.aar
*.ap_
*.aab

# Files for the ART/Dalvik VM
*.dex

# Java class files
*.class

# Generated files
bin/
gen/
out/
#  Uncomment the following line in case you need and you don't have the release build type files in your app
# release/

# Gradle files
.gradle/
build/

# Local configuration file (sdk path, etc)
local.properties

# Proguard folder generated by Eclipse
proguard/

# Log Files
*.log

# Android Studio Navigation editor temp files
.navigation/

# Android Studio captures folder
captures/

# IntelliJ
*.iml
.idea/
.idea/workspace.xml
.idea/tasks.xml
.idea/gradle.xml
.idea/assetWizardSettings.xml
.idea/dictionaries
.idea/libraries
# Android Studio 3 in .gitignore file.
.idea/caches
.idea/modules.xml
# Comment next line if keeping position of elements in Navigation Editor is relevant for you
.idea/navEditor.xml

# Keystore files
# Uncomment the following lines if you do not want to check your keystore files in.
#*.jks
#*.keystore

# External native build folder generated in Android Studio 2.2 and later
.externalNativeBuild
.cxx/

# Google Services (e.g. APIs or Firebase)
# google-services.json

# Freeline
freeline.py
freeline/
freeline_project_description.json

# fastlane
fastlane/report.xml
fastlane/Preview.html
fastlane/screenshots
fastlane/test_output
fastlane/readme.md

# Version control
vcs.xml

# lint
lint/intermediates/
lint/generated/
lint/outputs/
lint/tmp/
# lint/reports/

# Android Profiling
*.hprof
"""



# Templates
adds = {
    "di.py": {
        "type": "file",
        "module": "db-oracle",
        "path": "src/di.py",
        "code": {
            "imports": [
                "from zpy.utils import get_env_or_throw as var\n",
                "from zdb.oracle import ZOracle, ZDBConfig\n"
            ],
            "blocks": [
                "\n",
                "db_config: ZDBConfig = ZDBConfig(\"DB_USER\", \"DB_PASSWORD\", \"DB_NAME\",\"DB_HOST\", 1521,service=\"XE\")\n"
                "db_mngr: ZOracle = ZOracle.setup_of(db_config)\n"
                "\n",
                "if var('ENVIRONMENT') == 'local':\n",
                "   # Setup only the environment is local.\n",
                "   db_mngr.init_local_client(path= var('ORACLE_CLIENT_PATH'))\n",
                "\n"
            ]
        }
    },
    "context_imports": {
        "type": "file",
        "module": "db-oracle",
        "path": "src/di.py",
        "code": {
            "imports": [
                "\n"
                "# Imports of @context_name context 📦\n"
                "from contexts.@context_name import @usecase_class, @repository_name, @repository_impl\n"
            ],
            "blocks": [
                "\n# Creation of instances of use cases, repositories etc. of the context: @context_name 📦\n",
                "@context_name_repository: @repository_name = @repository_impl()\n"
                "@usecase_var_uc: UseCase[Any, Any] = @usecase_class(@context_name_repository)\n"
                "\n"
            ]
        }
    },
    "use-cases": {
        "type": "file",
        "module": "db-oracle",
        "path": "src/di.py",
        "code": {
            "imports": [
                "\n"
                "# Import use case of @context_name context 📦\n"
                "from contexts.@context_name.application.@usecase_file import @usecase_class\n"
            ],
            "blocks": [
                "\n# Creation instance of use cases of the context: @context_name 📦\n",
                "@usecase_var: UseCase[Any, Any] = @usecase_class(@context_name_repository)\n"
                "\n"
            ]
        }
    }
}



from enum import Enum
from typing import Any, List, Optional, Tuple, Callable
from requests import post, Response, get, patch, delete, put
from zpy.logger import zL
from datetime import timedelta


class HLevels(Enum):
    NOTHING = -1
    URL = 1
    HEADERS = 2
    PARAMS = 3
    PAYLOAD = 4


def zhl_write(text: str, value: Any):
    def wrapper():
        if value is not None:
            zL.i(f"{text} {value}")

    return wrapper


class ZHttp:
    """HTTP Wrapper


    Raises:
        ValueError: [description]
    """

    global_options: dict = {"BASE_URL": None, "LOG_LEVEL": HLevels.NOTHING}

    @staticmethod
    def setup(base: str, log_level: HLevels) -> None:
        ZHttp.global_options["BASE_URL"] = base
        ZHttp.global_options["LOG_LEVEL"] = log_level

    @staticmethod
    def __prepare_request__(
            url, path, params, headers, data, json, log_level
    ) -> Tuple[str, HLevels]:
        final_url: str = ZHttp.global_options["BASE_URL"]
        if url is not None and path is not None:
            final_url = f"{url}{path}"
        if url is not None and path is None:
            final_url = url
        if url is None and path is not None:
            final_url = f"{final_url}{path}"

        if final_url is None:
            raise ValueError("URL not configured!")
        real_log_level = ZHttp.global_options["LOG_LEVEL"]
        try:
            data_to_log: List[Any] = [
                zhl_write(f"Start HTTP [POST] -", final_url),
                zhl_write(f"Params:", params),
                zhl_write(f"Headers:", headers),
                zhl_write(f"Body:", data if json is None else json),
            ]

            if log_level is not None:
                real_log_level = log_level

            for i in range(0, real_log_level.value):
                data_to_log[i]()
        except Exception as e:
            zL.ex("An error occurred while logging http request.")
        return final_url, real_log_level

    @staticmethod
    def __logging_response__(
            result: Response, final_url, log_response_headers, real_log_level
    ):
        parsed = result.json()
        response_logs = [
            zhl_write(f"End Http [POST] - {result.status_code}", final_url),
            zhl_write(
                f"Elapsed: {str(timedelta(seconds=result.elapsed.seconds))}",
                "",
            ),
            zhl_write(
                f"Headers:",
                dict(zip(result.headers.keys(), result.headers.values()))
                if log_response_headers is True
                else None,
            ),
            zhl_write(f"Body:", parsed),
        ]
        for i in range(0, real_log_level.value):
            response_logs[i]()

    @staticmethod
    def get(
            url: Optional[str] = None,
            path: Optional[str] = None,
            params: Any = None,
            data: Any = None,
            headers: Any = None,
            cookies: Any = None,
            files: Any = None,
            auth: Any = None,
            timeout: Any = None,
            allow_redirects: bool = None,
            proxies: Any = None,
            hooks: Any = None,
            stream: Any = None,
            verify: Any = None,
            cert: Any = None,
            json: Any = None,
            log_level: HLevels = None,
            log_response_headers: bool = False,
            control_failure: bool = False,
    ):
        final_url, real_log_level = ZHttp.__prepare_request__(
            url, path, params, headers, data, json, log_level
        )
        try:
            result: Response = get(
                url=final_url,
                json=json,
                data=data,
                params=params,
                headers=headers,
                cookies=cookies,
                files=files,
                auth=auth,
                timeout=timeout,
                allow_redirects=allow_redirects,
                proxies=proxies,
                hooks=hooks,
                stream=stream,
                verify=verify,
                cert=cert,
            )
            if real_log_level != HLevels.NOTHING:
                ZHttp.__logging_response__(
                    result, final_url, log_response_headers, real_log_level
                )
            return result
        except Exception as e:
            zL.ex("The http request failed...")
            if control_failure is False:
                raise e
        return None

    @staticmethod
    def post(
            url: Optional[str] = None,
            path: Optional[str] = None,
            params: Any = None,
            data: Any = None,
            headers: Any = None,
            cookies: Any = None,
            files: Any = None,
            auth: Any = None,
            timeout: Any = None,
            allow_redirects: bool = None,
            proxies: Any = None,
            hooks: Any = None,
            stream: Any = None,
            verify: Any = None,
            cert: Any = None,
            json: Any = None,
            log_level: HLevels = None,
            log_response_headers: bool = False,
            control_failure: bool = False,
    ) -> Optional[Response]:
        final_url, real_log_level = ZHttp.__prepare_request__(
            url, path, params, headers, data, json, log_level
        )
        try:
            result: Response = post(
                url=final_url,
                json=json,
                data=data,
                params=params,
                headers=headers,
                cookies=cookies,
                files=files,
                auth=auth,
                timeout=timeout,
                allow_redirects=allow_redirects,
                proxies=proxies,
                hooks=hooks,
                stream=stream,
                verify=verify,
                cert=cert,
            )
            if real_log_level != HLevels.NOTHING:
                ZHttp.__logging_response__(
                    result, final_url, log_response_headers, real_log_level
                )
            return result
        except Exception as e:
            zL.ex("The http request failed...")
            if control_failure is False:
                raise e
        return None

    @staticmethod
    def put(
            url: Optional[str] = None,
            path: Optional[str] = None,
            params: Any = None,
            data: Any = None,
            headers: Any = None,
            cookies: Any = None,
            files: Any = None,
            auth: Any = None,
            timeout: Any = None,
            allow_redirects: bool = None,
            proxies: Any = None,
            hooks: Any = None,
            stream: Any = None,
            verify: Any = None,
            cert: Any = None,
            json: Any = None,
            log_level: HLevels = None,
            log_response_headers: bool = False,
            control_failure: bool = False,
    ) -> Optional[Response]:
        final_url, real_log_level = ZHttp.__prepare_request__(
            url, path, params, headers, data, json, log_level
        )
        try:
            result: Response = put(
                url=final_url,
                json=json,
                data=data,
                params=params,
                headers=headers,
                cookies=cookies,
                files=files,
                auth=auth,
                timeout=timeout,
                allow_redirects=allow_redirects,
                proxies=proxies,
                hooks=hooks,
                stream=stream,
                verify=verify,
                cert=cert,
            )
            if real_log_level != HLevels.NOTHING:
                ZHttp.__logging_response__(
                    result, final_url, log_response_headers, real_log_level
                )
            return result
        except Exception as e:
            zL.ex("The http request failed...")
            if control_failure is False:
                raise e
        return None

    @staticmethod
    def patch(
            url: Optional[str] = None,
            path: Optional[str] = None,
            params: Any = None,
            data: Any = None,
            headers: Any = None,
            cookies: Any = None,
            files: Any = None,
            auth: Any = None,
            timeout: Any = None,
            allow_redirects: bool = None,
            proxies: Any = None,
            hooks: Any = None,
            stream: Any = None,
            verify: Any = None,
            cert: Any = None,
            json: Any = None,
            log_level: HLevels = None,
            log_response_headers: bool = False,
            control_failure: bool = False,
    ) -> Optional[Response]:
        final_url, real_log_level = ZHttp.__prepare_request__(
            url, path, params, headers, data, json, log_level
        )
        try:
            result: Response = patch(
                url=final_url,
                json=json,
                data=data,
                params=params,
                headers=headers,
                cookies=cookies,
                files=files,
                auth=auth,
                timeout=timeout,
                allow_redirects=allow_redirects,
                proxies=proxies,
                hooks=hooks,
                stream=stream,
                verify=verify,
                cert=cert,
            )
            if real_log_level != HLevels.NOTHING:
                ZHttp.__logging_response__(
                    result, final_url, log_response_headers, real_log_level
                )
            return result
        except Exception as e:
            zL.ex("The http request failed...")
            if control_failure is False:
                raise e
        return None

    @staticmethod
    def delete(
            url: Optional[str] = None,
            path: Optional[str] = None,
            params: Any = None,
            data: Any = None,
            headers: Any = None,
            cookies: Any = None,
            files: Any = None,
            auth: Any = None,
            timeout: Any = None,
            allow_redirects: bool = None,
            proxies: Any = None,
            hooks: Any = None,
            stream: Any = None,
            verify: Any = None,
            cert: Any = None,
            json: Any = None,
            log_level: HLevels = None,
            log_response_headers: bool = False,
            control_failure: bool = False,
    ) -> Optional[Response]:
        final_url, real_log_level = ZHttp.__prepare_request__(
            url, path, params, headers, data, json, log_level
        )
        try:
            result: Response = delete(
                url=final_url,
                json=json,
                data=data,
                params=params,
                headers=headers,
                cookies=cookies,
                files=files,
                auth=auth,
                timeout=timeout,
                allow_redirects=allow_redirects,
                proxies=proxies,
                hooks=hooks,
                stream=stream,
                verify=verify,
                cert=cert,
            )
            if real_log_level != HLevels.NOTHING:
                ZHttp.__logging_response__(
                    result, final_url, log_response_headers, real_log_level
                )
            return result
        except Exception as e:
            zL.ex("The http request failed...")
            if control_failure is False:
                raise e
        return None


class ZHttpResponse(object):
    """
    Wrapper
    """

    def __init__(self, response: Response):
        self.raw = response

    def json(self) -> dict:
        return self.raw.json()

    def status_is(self, status: int) -> bool:
        return False if self.raw is None else self.raw.status_code == status

    def status_is_and(self, status: int, action: Callable[[Response], Optional[Any]]) -> Optional[Any]:
        """

        @param status:
        @param action:
        @return:
        """
        if self.raw is not None and self.raw.status_code == status:
            return action(self.raw)
        return None

    def is_ok(self) -> bool:
        """

        @return: http status response is success
        """
        return False if self.raw is None else self.raw.ok



# Created by Noé Cruz | Zurckz 22 at 20/03/2022
# See https://www.linkedin.com/in/zurckz

import os
import sys
import json
from pathlib import Path


def read_event(name: str = None, event_file_path: str = None):
    event_path = f"{os.getcwd()}\\events\\{name}" if name else event_file_path
    if not event_path:
        raise ValueError("Path file or name is required...")
    with open(event_path, 'r') as sm:
        return json.loads(sm.read())


def add_source_to_path(src_dir: str = None, replacement: str = None):
    current_dir = src_dir
    if not current_dir:
        current_dir = os.getcwd()
        current_dir = current_dir.replace('\\tests' if not replacement else replacement, '\\src')
    sys.path.append(current_dir)
    path = Path(current_dir)
    sys.path.append(str(path.parent.absolute()))



import logging
from typing import Callable, Any, Optional
from zpy.containers import shared_container
from zpy.api.http.response import _useAwsRequestId
from zpy.logger import zL
from zpy.utils.values import if_null_get
from timeit import default_timer as timer
from datetime import timedelta
from functools import wraps
from zpy.app import zapp_context as api


def store_request_id(context) -> Optional[str]:
    """Extract aws request id from context

    Args:
        context ([type]): Lambda context
    """
    try:
        shared_container["aws_request_id"] = context.aws_request_id
        return context.aws_request_id
    except Exception as e:
        zL.ex("An error occurred while extracting aws request id", exc_info=e)
        return None


def event_processors(storage: dict, use_id: bool, logs: bool, send_logs: bool, *args, **kwargs):
    """Lambda event processors

    Args:
        @param storage:
        @param use_id:
        @param logs:
        @param send_logs:
    """
    try:
        if len(args) >= 2:
            event = args[0]
            storage['request'] = event
            if logs:
                api().logger.info(f"Request: {event}", shippable=send_logs)
            if _useAwsRequestId or use_id:
                storage['request_id'] = store_request_id(args[1])

        else:
            if "event" in kwargs:
                storage['request'] = kwargs['event']
                if logs:
                    api().logger.info(f"Request: {kwargs['event']}", shippable=send_logs)
            if "context" in kwargs:
                if _useAwsRequestId or use_id:
                    storage['request_id'] = store_request_id(args[1])
    except Exception as e:
        api().logger.ex("An error occurred while processing event!", exc_info=e)


def aws_lambda(logs: bool = True, save_id: bool = False, measure_time: bool = True, send_logs: bool = False,
               event_sender: Optional[Callable[[dict], Any]] = None):
    """Lambda Handler

    Args:
        @param event_sender:
        @param logs: (bool, optional): Logging request and response. Defaults to False.
        @param save_id: (bool, optional): Register aws lambda request id. Defaults to True.
        @param measure_time: (bool, optional): Measure elapsed execution time. Defaults to True.
        @param send_logs: Send event logs by log sender configured
    """
    api().release_logger()
    event = {}

    def callable_fn(invoker: Callable):
        @wraps(invoker)
        def wrapper(*args, **kwargs):
            event_processors(event, save_id, logs, send_logs, *args, **kwargs)
            start = 0.0
            if if_null_get(measure_time, False):
                start = timer()
            result = invoker(*args, **kwargs)
            event['response'] = result
            if logs:
                api().logger.info(f"Response: {result}", shippable=send_logs)
            if if_null_get(measure_time, False):
                end = timer()
                api().logger.info(f"Elapsed execution time: {timedelta(seconds=end - start)}", shippable=send_logs)
            if event_sender:
                event_sender(event)
            return result

        return wrapper

    return callable_fn
